use std::io::{Read, BufRead};

pub trait Peek {
    type Item;
    fn peek(&mut self) -> Option<&Self::Item> {
        self.peek_n(1)?.first()
    }

    fn peek_n(&mut self, n: usize) -> Option<&[Self::Item]>;
}

pub struct Reader<R> {
    inner: R
}

impl<R> Reader<R> {
    pub fn new(inner: R) -> Self {
        Self { inner }
    }
}

impl<R: BufRead> Peek for Reader<R> {
    type Item = u8;

    fn peek_n(&mut self, n: usize) -> Option<&[Self::Item]> {
        Some(&self.fill_buf().ok()?[..n])
    }
}

impl<R: Read> Iterator for Reader<R> {
    type Item = u8;

    fn next(&mut self) -> Option<Self::Item> {
        let mut buf = [0u8];
        self.inner.read(&mut buf).ok()?;
        Some(buf[0])
    }
}

impl<R: Read> Read for Reader<R> {
    fn read(&mut self, buf: &mut [u8]) -> std::io::Result<usize> {
        self.inner.read(buf)
    }
}

impl<R: BufRead> BufRead for Reader<R> {
    fn fill_buf(&mut self) -> std::io::Result<&[u8]> {
        self.inner.fill_buf()
    }

    fn consume(&mut self, amt: usize) {
        self.inner.consume(amt)
    }
}

#[cfg(test)]
mod tests {
    use std::io::BufReader;

    use crate::reader::Peek;

    use super::Reader;

    #[test]
    fn basic_read() {
        let mut reader = Reader::new(BufReader::new(b"this is a test string".as_slice()));
        assert_eq!(Some(b't'), reader.next());
        assert_eq!(Some(b'h'), reader.next());
        assert_eq!(Some(b'i'), reader.next());
        assert_eq!(Some(b's'), reader.next());
    }

    #[test]
    fn basic_peek() {
        let mut reader = Reader::new(BufReader::new(b"this is a test string".as_slice()));
        assert_eq!(Some(&b't'), reader.peek());
        assert_eq!(Some(&b't'), reader.peek());
        assert_eq!("this", String::from_utf8_lossy(reader.peek_n(4).unwrap()));
        assert_eq!("this", String::from_utf8_lossy(reader.peek_n(4).unwrap()));
    }
}

